def configuration(parent_package="", top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration("cesium", parent_package, top_path)
    config.add_subpackage("features")
    config.add_data_dir("tests")
    config.add_data_dir("tests/data")
    config.add_data_dir("features/tests/data")
    return config


if __name__ == "__main__":
    from numpy.distutils.core import setup

    config = configuration(top_path="").todict()
    setup(**config)



import numpy as np
import scipy.stats as stats
from ._lomb_scargle import lomb_scargle


def lomb_scargle_model(
    time, signal, error, sys_err=0.05, nharm=8, nfreq=3, tone_control=5.0
):
    """Simultaneous fit of a sum of sinusoids by weighted least squares:
           y(t) = Sum_k Ck*t^k + Sum_i Sum_j A_ij sin(2*pi*j*fi*(t-t0)+phi_j),
           i=[1,nfreq], j=[1,nharm]

    Parameters
    ----------
    time : array_like
        Array containing time values.

    signal : array_like
        Array containing data values.

    error : array_like
        Array containing measurement error values.

    nharm : int
        Number of harmonics to fit for each frequency.

    nfreq : int
        Number of frequencies to fit.

    Returns
    -------
    dict
        Dictionary containing fitted parameter values. Parameters specific to
        a specific fitted frequency are stored in a list of dicts at
        model_dict['freq_fits'], each of which contains the output of
        fit_lomb_scargle(...)

    """

    dy0 = np.sqrt(error**2 + sys_err**2)

    wt = 1.0 / dy0**2
    time = time.copy() - min(time)  # speeds up lomb_scargle code to have min(time)==0
    signal = signal.copy()

    chi0 = np.dot(signal**2, wt)

    # TODO parametrize?
    f0 = 1.0 / max(time)
    df = 0.8 / max(time)  # 20120202 :    0.1/Xmax
    fmax = 33.0  # pre 20120126: 10. # 25
    numf = int((fmax - f0) / df)  # TODO !!! this is off by 1 point, fix?

    model_dict = {"freq_fits": []}
    lambda0_range = [
        -np.log10(len(time)),
        8,
    ]  # these numbers "fix" the strange-amplitude effect
    for i in range(nfreq):
        if i == 0:
            fit = fit_lomb_scargle(
                time,
                signal,
                dy0,
                f0,
                df,
                numf,
                tone_control=tone_control,
                lambda0_range=lambda0_range,
                nharm=nharm,
                detrend_order=1,
            )
            model_dict["trend"] = fit["trend_coef"][1]
        else:
            fit = fit_lomb_scargle(
                time,
                signal,
                dy0,
                f0,
                df,
                numf,
                tone_control=tone_control,
                lambda0_range=lambda0_range,
                nharm=nharm,
                detrend_order=0,
            )
        model_dict["freq_fits"].append(fit)
        signal -= fit["model"]
        model_dict["freq_fits"][-1]["resid"] = signal.copy()
        if i == 0:
            model_dict["varrat"] = np.dot(signal**2, wt) / chi0

    model_dict["nfreq"] = nfreq
    model_dict["nharm"] = nharm
    model_dict["chi2"] = fit["chi2"]
    model_dict["f0"] = f0
    model_dict["df"] = df
    model_dict["numf"] = numf

    return model_dict


def lprob2sigma(lprob):
    """Translate a log_e(probability) to units of Gaussian sigmas."""
    if lprob > -36.0:
        sigma = stats.norm.ppf(1.0 - 0.5 * np.exp(lprob))
    else:
        sigma = np.sqrt(np.log(2.0 / np.pi) - 2.0 * np.log(8.2) - 2.0 * lprob)
        f = 0.5 * np.log(2.0 / np.pi) - 0.5 * sigma**2 - np.log(sigma) - lprob
        sigma += f / (sigma + 1.0 / sigma)
    return sigma


def fit_lomb_scargle(
    time,
    signal,
    error,
    f0,
    df,
    numf,
    nharm=8,
    psdmin=6.0,
    detrend_order=0,
    freq_zoom=10.0,
    tone_control=5.0,
    lambda0=1.0,
    lambda0_range=[-8, 6],
):
    """Calls C implementation of Lomb Scargle sinusoid fitting, which fits a
    single frequency with nharm harmonics to the data. Called repeatedly by
    lomb_scargle_model in order to produce a fit with multiple distinct
    frequencies.

    Inputs:
    time : array_like
        Array containing time values.

    signal : array_like
        Array containing data values.

    error : array_like
        Array containing measurement error values.

    f0 : float
        Smallest frequency value to consider.

    df : float
        Step size for frequency grid search.

    numf : int
        Number of frequencies for frequency grid search.

    nharm : int
        Number of harmonics to fit.

    detrend_order : int
        Order of polynomial detrending.

    psdmin : int
        Refine periodogram values with larger psd using multi-harmonic fit

    nharm : int
        Number of harmonics to use in refinement

    lambda0 : float
        Typical value for regularization parameter

    lambda0_range : [float, float]
        Allowable range for log10 of regularization parameter

    Returns
    -------
    dict
        Dictionary describing various parameters of the multiharmonic fit at
        the best-fit frequency
    """
    ntime = len(time)

    # For some reason we round this to the nearest even integer
    freq_zoom = round(freq_zoom / 2.0) * 2.0

    # Polynomial terms
    coef = np.zeros(detrend_order + 1, dtype="float64")
    norm = np.zeros(detrend_order + 1, dtype="float64")

    wth0 = 1.0 / error
    s0 = np.dot(wth0, wth0)
    wth0 /= np.sqrt(s0)

    cn = signal * wth0
    coef[0] = np.dot(cn, wth0)
    cn0 = coef[0]
    norm[0] = 1.0
    cn -= coef[0] * wth0
    vcn = 1.0

    # np.sin's and cosin's for later
    tt = 2.0 * np.pi * time
    sinx, cosx = np.sin(tt * f0) * wth0, np.cos(tt * f0) * wth0
    sinx_step, cosx_step = np.sin(tt * df), np.cos(tt * df)
    sinx_back, cosx_back = -np.sin(tt * df / 2.0), np.cos(tt * df / 2)
    sinx_smallstep, cosx_smallstep = np.sin(tt * df / freq_zoom), np.cos(
        tt * df / freq_zoom
    )

    npar = 2 * nharm
    hat_matr = np.zeros((npar, ntime), dtype="float64")
    hat0 = np.zeros((npar, detrend_order + 1), dtype="float64")
    hat_hat = np.zeros((npar, npar), dtype="float64")
    soln = np.zeros(npar, dtype="float64")
    psd = np.zeros(numf, dtype="float64")

    # Detrend the data and create the orthogonal detrending basis
    if detrend_order > 0:
        wth = np.zeros((detrend_order + 1, ntime), dtype="float64")
        wth[0, :] = wth0
    else:
        wth = wth0

    for i in range(detrend_order):
        f = wth[i, :] * tt / (2 * np.pi)
        for j in range(i + 1):
            f -= np.dot(f, wth[j, :]) * wth[j, :]
        norm[i + 1] = np.sqrt(np.dot(f, f))
        f /= norm[i + 1]
        coef[i + 1] = np.dot(cn, f)
        cn -= coef[i + 1] * f
        wth[i + 1, :] = f
        vcn += (f / wth0) ** 2

    chi0 = np.dot(cn, cn)
    varcn = chi0 / (ntime - 1 - detrend_order)
    psdmin *= 2 * varcn

    Tr = np.array(0.0, dtype="float64")
    ifreq = np.array(0, dtype="int32")
    lambda0 = np.array(lambda0 / s0, dtype="float64")
    lambda0_range = 10 ** np.array(lambda0_range, dtype="float64") / s0

    lomb_scargle(
        ntime,
        numf,
        nharm,
        detrend_order,
        psd,
        cn,
        wth,
        sinx,
        cosx,
        sinx_step,
        cosx_step,
        sinx_back,
        cosx_back,
        sinx_smallstep,
        cosx_smallstep,
        hat_matr,
        hat_hat,
        hat0,
        soln,
        chi0,
        freq_zoom,
        psdmin,
        tone_control,
        lambda0,
        lambda0_range,
        Tr,
        ifreq,
    )

    hat_hat /= s0
    ii = np.arange(nharm, dtype="int32")
    soln[0:nharm] /= (1.0 + ii) ** 2
    soln[nharm:] /= (1.0 + ii) ** 2
    hat_matr0 = np.outer(hat0[:, 0], wth0)
    for i in range(detrend_order):
        hat_matr0 += np.outer(hat0[:, i + 1], wth[i + 1, :])

    modl = np.dot(hat_matr.T, soln)
    coef0 = np.dot(soln, hat0)
    coef -= coef0
    hat_matr -= hat_matr0

    out_dict = {}
    out_dict["psd"] = psd
    out_dict["chi0"] = chi0 * s0
    if detrend_order > 0:
        out_dict["trend"] = np.dot(coef, wth) / wth0
    else:
        out_dict["trend"] = coef[0] + 0 * wth0
    out_dict["model"] = modl / wth0 + out_dict["trend"]

    j = psd.argmax()
    freq = f0 + df * j + (ifreq / freq_zoom - 1 / 2.0) * df
    tt = (time * freq) % 1.0
    out_dict["freq"] = freq
    out_dict["s0"] = s0
    out_dict["chi2"] = (chi0 - psd[j]) * s0
    out_dict["psd"] = psd[j] * 0.5 / varcn
    out_dict["lambda"] = lambda0 * s0
    #    out_dict['gcv_weight'] = (1 - 3. / ntime) / Tr
    out_dict["trace"] = Tr
    out_dict["nu0"] = ntime - npar
    npars = (1 - Tr) * ntime / 2.0
    out_dict["nu"] = ntime - npars
    out_dict["npars"] = npars

    A0, B0 = soln[0:nharm], soln[nharm:]
    hat_hat /= np.outer(
        np.hstack(((1.0 + ii) ** 2, (1.0 + ii) ** 2)),
        np.hstack(((1.0 + ii) ** 2, (1.0 + ii) ** 2)),
    )
    err2 = np.diag(hat_hat)
    vA0, vB0 = err2[0:nharm], err2[nharm:]
    covA0B0 = hat_hat[(ii, nharm + ii)]

    vmodl = vcn / s0 + np.dot((hat_matr / wth0).T, np.dot(hat_hat, hat_matr / wth0))
    vmodl0 = vcn / s0 + np.dot((hat_matr0 / wth0).T, np.dot(hat_hat, hat_matr0 / wth0))
    out_dict["model_error"] = np.sqrt(np.diag(vmodl))
    out_dict["trend_error"] = np.sqrt(np.diag(vmodl0))

    amp = np.sqrt(A0**2 + B0**2)
    damp = np.sqrt(A0**2 * vA0 + B0**2 * vB0 + 2.0 * A0 * B0 * covA0B0) / amp
    phase = np.arctan2(B0, A0)
    rel_phase = phase - phase[0] * (1.0 + ii)
    rel_phase = np.arctan2(np.sin(rel_phase), np.cos(rel_phase))
    dphase = 0.0 * rel_phase
    for i in range(nharm - 1):
        j = i + 1
        v = np.array(
            [
                -A0[0] * (1.0 + j) / amp[0] ** 2,
                B0[0] * (1.0 + j) / amp[0] ** 2,
                A0[j] / amp[j] ** 2,
                -B0[j] / amp[j] ** 2,
            ]
        )
        jj = np.array([0, nharm, j, j + nharm])
        m = hat_hat[np.ix_(jj, jj)]
        dphase[j] = np.sqrt(np.dot(np.dot(v, m), v))

    out_dict["amplitude"] = amp
    out_dict["amplitude_error"] = damp
    out_dict["rel_phase"] = rel_phase
    out_dict["rel_phase_error"] = dphase
    out_dict["time0"] = -phase[0] / (2 * np.pi * freq)

    ncp = norm.cumprod()
    out_dict["trend_coef"] = coef / ncp
    out_dict["y_offset"] = out_dict["trend_coef"][0] - cn0

    prob = stats.f.sf(
        0.5
        * (ntime - 1.0 - detrend_order)
        * (1.0 - out_dict["chi2"] / out_dict["chi0"]),
        2,
        ntime - 1 - detrend_order,
    )
    out_dict["signif"] = lprob2sigma(np.log(prob))

    return out_dict


def get_lomb_frequency(lomb_model, i):
    """Get the ith frequency from a fitted Lomb-Scargle model."""
    return lomb_model["freq_fits"][i - 1]["freq"]


def get_lomb_amplitude(lomb_model, i, j):
    """
    Get the amplitude of the jth harmonic of the ith frequency from a fitted
    Lomb-Scargle model.
    """
    return lomb_model["freq_fits"][i - 1]["amplitude"][j - 1]


def get_lomb_rel_phase(lomb_model, i, j):
    """
    Get the relative phase of the jth harmonic of the ith frequency from a
    fitted Lomb-Scargle model.
    """
    return lomb_model["freq_fits"][i - 1]["rel_phase"][j - 1]


def get_lomb_amplitude_ratio(lomb_model, i):
    """
    Get the ratio of the amplitudes of the first harmonic for the ith and first
    frequencies from a fitted Lomb-Scargle model.
    """
    return (
        lomb_model["freq_fits"][i - 1]["amplitude"][0]
        / lomb_model["freq_fits"][0]["amplitude"][0]
    )


def get_lomb_frequency_ratio(lomb_model, i):
    """
    Get the ratio of the ith and first frequencies from a fitted Lomb-Scargle
    model.
    """
    return lomb_model["freq_fits"][i - 1]["freq"] / lomb_model["freq_fits"][0]["freq"]


def get_lomb_signif_ratio(lomb_model, i):
    """
    Get the ratio of the significances (in sigmas) of the ith and first
    frequencies from a fitted Lomb-Scargle model.
    """
    return (
        lomb_model["freq_fits"][i - 1]["signif"] / lomb_model["freq_fits"][0]["signif"]
    )


def get_lomb_lambda(lomb_model):
    """Get the regularization parameter of a fitted Lomb-Scargle model."""
    return lomb_model["freq_fits"][0]["lambda"]


def get_lomb_signif(lomb_model):
    """
    Get the significance (in sigmas) of the first frequency from a fitted
    Lomb-Scargle model.
    """
    return lomb_model["freq_fits"][0]["signif"]


def get_lomb_varrat(lomb_model):
    """
    Get the fraction of the variance explained by the first frequency of a
    fitted Lomb-Scargle model.
    """
    return lomb_model["varrat"]


def get_lomb_trend(lomb_model):
    """Get the linear trend of a fitted Lomb-Scargle model."""
    return lomb_model["trend"]


def get_lomb_y_offset(lomb_model):
    """Get the y-intercept of a fitted Lomb-Scargle model."""
    return lomb_model["freq_fits"][0]["y_offset"]



from tabulate import tabulate
import re
from cesium.features.graphs import (
    feature_categories,
    dask_feature_graph,
    extra_feature_docs,
)


def feature_graph_to_rst_table(graph, category_name):
    """Convert feature graph to Sphinx-compatible ReST table."""
    header = [category_name, "Description"]
    table = []
    for feature_name in sorted(
        graph,
        key=lambda s: [int(t) if t.isdigit() else t for t in re.split(r"(\d+)", s)],
    ):
        description = (
            extra_feature_docs[feature_name]
            if feature_name in extra_feature_docs
            else " ".join(
                [
                    e.strip()
                    for e in graph[feature_name][0]
                    .__doc__.split("\n\n")[0]
                    .strip()
                    .split("\n")
                ]
            )
        )
        table.append([feature_name, description])

    return tabulate(table, headers=header, tablefmt="rst")


def write_feature_tables(fname):
    with open(fname, "w") as f:
        f.write(
            "==============================\n"
            "Cesium Features - By Category\n"
            "==============================\n\n"
        )

        dfg = dask_feature_graph

        for category in feature_categories:
            graph = {feature: dfg[feature] for feature in feature_categories[category]}

            f.write(feature_graph_to_rst_table(graph, category) + "\n\n")



import os
from os.path import join as pjoin
import numpy as np
import pandas as pd
import scipy.stats
import dask

from cesium import featurize
from cesium.tests.fixtures import sample_values, sample_ts_files, sample_featureset

import numpy.testing as npt
import pytest


DATA_PATH = pjoin(os.path.dirname(__file__), "data")
FEATURES_CSV_PATH = pjoin(DATA_PATH, "test_features_with_targets.csv")


def test_featurize_files_function(tmpdir):
    """Test featurize function for on-disk time series"""
    with sample_ts_files(size=4, labels=["A", "B"]) as ts_paths:
        fset, labels = featurize.featurize_ts_files(
            ts_paths, features_to_use=["std_err"], scheduler=dask.get
        )
    assert "std_err" in fset
    assert fset.shape == (4, 1)
    npt.assert_array_equal(labels, ["A", "B", "A", "B"])


def test_featurize_time_series_single():
    """Test featurize wrapper function for single time series"""
    t, m, e = sample_values()
    features_to_use = ["amplitude", "std_err"]
    meta_features = {"meta1": 0.5}
    fset = featurize.featurize_time_series(
        t, m, e, features_to_use, meta_features, scheduler=dask.get
    )
    assert fset["amplitude"].values.dtype == np.float64


def test_featurize_time_series_single_multichannel():
    """Test featurize wrapper function for single multichannel time series"""
    n_channels = 3
    t, m, e = sample_values(channels=n_channels)
    features_to_use = ["amplitude", "std_err"]
    meta_features = {"meta1": 0.5}
    fset = featurize.featurize_time_series(
        t, m, e, features_to_use, meta_features, scheduler=dask.get
    )
    assert ("amplitude", 0) in fset.columns
    assert "meta1" in fset.columns


def test_featurize_time_series_multiple():
    """Test featurize wrapper function for multiple time series"""
    n_series = 5
    list_of_series = [sample_values() for i in range(n_series)]
    times, values, errors = [list(x) for x in zip(*list_of_series)]
    features_to_use = ["amplitude", "std_err"]
    meta_features = [{"meta1": 0.5}] * n_series
    fset = featurize.featurize_time_series(
        times, values, errors, features_to_use, meta_features, scheduler=dask.get
    )
    npt.assert_array_equal(
        sorted(fset.columns.get_level_values("feature")),
        ["amplitude", "meta1", "std_err"],
    )


def test_featurize_time_series_multiple_multichannel():
    """Test featurize wrapper function for multiple multichannel time series"""
    n_series = 5
    n_channels = 3
    list_of_series = [sample_values(channels=n_channels) for i in range(n_series)]
    times, values, errors = [list(x) for x in zip(*list_of_series)]
    features_to_use = ["amplitude", "std_err"]
    meta_features = {"meta1": 0.5}
    fset = featurize.featurize_time_series(
        times, values, errors, features_to_use, meta_features, scheduler=dask.get
    )
    assert ("amplitude", 0) in fset.columns
    assert "meta1" in fset.columns


def test_featurize_time_series_uneven_multichannel():
    """Test featurize wrapper function for uneven-length multichannel data"""
    n_channels = 3
    t, m, e = sample_values(channels=n_channels)
    t = [[t, t[0:-5], t[0:-10]]]
    m = [[m[0], m[1][0:-5], m[2][0:-10]]]
    e = [[e[0], e[1][0:-5], e[2][0:-10]]]
    features_to_use = ["amplitude", "std_err"]
    meta_features = {"meta1": 0.5}
    fset = featurize.featurize_time_series(
        t, m, e, features_to_use, meta_features, scheduler=dask.get
    )
    assert ("amplitude", 0) in fset.columns
    assert "meta1" in fset.columns


def test_featurize_time_series_custom_functions():
    """Test featurize wrapper function for time series w/ custom functions"""
    n_channels = 3
    t, m, e = sample_values(channels=n_channels)
    features_to_use = ["amplitude", "std_err", "test_f"]
    meta_features = {"meta1": 0.5}
    custom_functions = {"test_f": lambda t, m, e: np.pi}
    fset = featurize.featurize_time_series(
        t,
        m,
        e,
        features_to_use,
        meta_features,
        custom_functions=custom_functions,
        scheduler=dask.get,
    )
    npt.assert_array_equal(fset["test_f", 0], np.pi)
    assert ("amplitude", 0) in fset.columns
    assert "meta1" in fset.columns


def test_featurize_time_series_custom_dask_graph():
    """Test featurize wrapper function for time series w/ custom dask graph"""
    n_channels = 3
    t, m, e = sample_values(channels=n_channels)
    features_to_use = ["amplitude", "std_err", "test_f", "test_meta"]
    meta_features = {"meta1": 0.5}
    custom_functions = {
        "test_f": (lambda x: x.min() - x.max(), "amplitude"),
        "test_meta": (lambda x: 2.0 * x, "meta1"),
    }
    fset = featurize.featurize_time_series(
        t,
        m,
        e,
        features_to_use,
        meta_features,
        custom_functions=custom_functions,
        scheduler=dask.get,
    )
    assert ("amplitude", 0) in fset.columns
    assert ("test_f", 0) in fset.columns
    assert ("test_meta", 0) in fset.columns


def test_featurize_time_series_default_times():
    """Test featurize wrapper function for time series w/ missing times"""
    n_channels = 3
    _, m, e = sample_values(channels=n_channels)
    features_to_use = ["amplitude", "std_err"]
    meta_features = {}
    fset = featurize.featurize_time_series(
        None, m, e, features_to_use, meta_features, scheduler=dask.get
    )

    m = [[m[0], m[1][0:-5], m[2][0:-10]]]
    e = [[e[0], e[1][0:-5], e[2][0:-10]]]
    fset = featurize.featurize_time_series(
        None, m, e, features_to_use, meta_features, scheduler=dask.get
    )

    m = m[0][0]
    e = e[0][0]
    fset = featurize.featurize_time_series(
        None, m, e, features_to_use, meta_features, scheduler=dask.get
    )
    assert ("amplitude", 0) in fset.columns


def test_featurize_time_series_default_errors():
    """Test featurize wrapper function for time series w/ missing errors"""
    n_channels = 3
    t, m, _ = sample_values(channels=n_channels)
    features_to_use = ["amplitude", "std_err"]
    meta_features = {}
    fset = featurize.featurize_time_series(
        t, m, None, features_to_use, meta_features, scheduler=dask.get
    )

    t = [[t, t[0:-5], t[0:-10]]]
    m = [[m[0], m[1][0:-5], m[2][0:-10]]]
    fset = featurize.featurize_time_series(
        t, m, None, features_to_use, meta_features, scheduler=dask.get
    )

    t = t[0][0]
    m = m[0][0]
    fset = featurize.featurize_time_series(
        t, m, None, features_to_use, meta_features, scheduler=dask.get
    )
    assert ("amplitude", 0) in fset.columns


def test_featurize_time_series_pandas_metafeatures():
    """Test featurize function for metafeatures passed as Series/DataFrames."""
    t, m, e = sample_values()
    features_to_use = ["amplitude", "std_err"]
    meta_features = pd.Series({"meta1": 0.5})
    fset = featurize.featurize_time_series(
        t, m, e, features_to_use, meta_features, scheduler=dask.get
    )
    npt.assert_allclose(fset["meta1"], 0.5)

    n_series = 5
    list_of_series = [sample_values() for i in range(n_series)]
    times, values, errors = [list(x) for x in zip(*list_of_series)]
    features_to_use = ["amplitude", "std_err"]
    meta_features = pd.DataFrame({"meta1": [0.5] * n_series, "meta2": [0.8] * n_series})
    fset = featurize.featurize_time_series(
        times, values, errors, features_to_use, meta_features, scheduler=dask.get
    )
    npt.assert_allclose(fset["meta1"], 0.5)
    npt.assert_allclose(fset["meta2"], 0.8)


def test_impute():
    """Test imputation of missing Featureset values."""
    fset, labels = sample_featureset(
        5,
        1,
        ["amplitude"],
        ["class1", "class2"],
        names=["a", "b", "c", "d", "e"],
        meta_features=["meta1"],
    )

    imputed = featurize.impute_featureset(fset)
    npt.assert_allclose(fset.amplitude.values, imputed.amplitude.values)
    assert isinstance(imputed, pd.DataFrame)

    fset.amplitude.values[0] = np.inf
    fset.amplitude.values[1] = np.nan
    amp_values = fset.amplitude.values[2:]
    other_values = fset.values.T.ravel()[2:]

    imputed = featurize.impute_featureset(fset, strategy="constant", value=None)
    npt.assert_allclose(
        -2 * np.nanmax(np.abs(other_values)), imputed.amplitude.values[0:2]
    )

    imputed = featurize.impute_featureset(fset, strategy="constant", value=-1e4)
    npt.assert_allclose(-1e4, imputed.amplitude.values[0:2])

    imputed = featurize.impute_featureset(fset, strategy="mean")
    npt.assert_allclose(np.mean(amp_values), imputed.amplitude.values[0:2])
    npt.assert_allclose(amp_values, imputed.amplitude.values[2:])

    imputed = featurize.impute_featureset(fset, strategy="median")
    npt.assert_allclose(np.median(amp_values), imputed.amplitude.values[0:2])
    npt.assert_allclose(amp_values, imputed.amplitude.values[2:])

    imputed = featurize.impute_featureset(fset, strategy="most_frequent")
    npt.assert_allclose(
        scipy.stats.mode(amp_values).mode.item(), imputed.amplitude.values[0:2]
    )
    npt.assert_allclose(amp_values, imputed.amplitude.values[2:])

    featurize.impute_featureset(fset, strategy="constant", value=-1e4, inplace=True)
    npt.assert_allclose(-1e4, fset.amplitude.values[0:2])

    with pytest.raises(NotImplementedError):
        featurize.impute_featureset(fset, strategy="blah")


def test_roundtrip_featureset(tmpdir):
    fset_path = os.path.join(str(tmpdir), "test.npz")
    for n_channels in [1, 3]:
        for labels in [["class1", "class2"], []]:
            fset, labels = sample_featureset(
                3,
                n_channels,
                ["amplitude"],
                labels,
                names=["a", "b", "c"],
                meta_features=["meta1"],
            )

            pred_probs = pd.DataFrame(
                np.random.random((len(fset), 2)),
                index=fset.index.values,
                columns=["class1", "class2"],
            )

            featurize.save_featureset(
                fset, fset_path, labels=labels, pred_probs=pred_probs
            )
            fset_loaded, data_loaded = featurize.load_featureset(fset_path)
            npt.assert_allclose(fset.values, fset_loaded.values)
            npt.assert_array_equal(fset.index, fset_loaded.index)
            npt.assert_array_equal(fset.columns, fset_loaded.columns)
            assert isinstance(fset_loaded, pd.DataFrame)
            npt.assert_array_equal(labels, data_loaded["labels"])
            npt.assert_allclose(pred_probs, data_loaded["pred_probs"])
            npt.assert_array_equal(
                pred_probs.columns, data_loaded["pred_probs"].columns
            )


def test_ignore_exceptions():
    import cesium.features.graphs

    def raise_exc(x):
        raise ValueError()

    old_value = cesium.features.graphs.dask_feature_graph["mean"]
    try:
        cesium.features.graphs.dask_feature_graph["mean"] = (raise_exc, "t")
        t, m, e = sample_values()
        features_to_use = ["mean"]
        with pytest.raises(ValueError):
            fset = featurize.featurize_time_series(
                t, m, e, features_to_use, scheduler=dask.get, raise_exceptions=True
            )
        fset = featurize.featurize_time_series(
            t, m, e, features_to_use, scheduler=dask.get, raise_exceptions=False
        )
        assert np.isnan(fset.values).all()
    finally:
        cesium.features.graphs.dask_feature_graph["mean"] = old_value



import numpy as np

from .cadence_features import (
    cad_prob,
    delta_t_hist,
    double_to_single_step,
    normalize_hist,
    find_sorted_peaks,
    peak_bin,
    peak_ratio,
)

from .common_functions import (
    maximum,
    median,
    max_slope,
    median_absolute_deviation,
    minimum,
    percent_beyond_1_std,
    percent_close_to_median,
    skew,
    std,
    weighted_average,
)
from .amplitude import (
    amplitude,
    percent_amplitude,
    flux_percentile_ratio,
    percent_difference_flux_percentile,
)
from .qso_model import qso_fit, get_qso_log_chi2_qsonu, get_qso_log_chi2nuNULL_chi2nu
from .stetson import stetson_j, stetson_k

from .lomb_scargle import (
    lomb_scargle_model,
    get_lomb_frequency,
    get_lomb_amplitude,
    get_lomb_rel_phase,
    get_lomb_amplitude_ratio,
    get_lomb_frequency_ratio,
    get_lomb_signif_ratio,
    get_lomb_lambda,
    get_lomb_signif,
    get_lomb_varrat,
    get_lomb_trend,
    get_lomb_y_offset,
)
from .lomb_scargle_fast import lomb_scargle_fast_period
from .num_alias import num_alias
from .periodic_model import (
    periodic_model,
    get_max_delta_mags,
    get_min_delta_mags,
    get_model_phi1_phi2,
)
from .period_folding import (
    period_folding,
    get_fold2P_slope_percentile,
    get_medperc90_2p_p,
    p2p_model,
    get_p2p_scatter_2praw,
    get_p2p_scatter_over_mad,
    get_p2p_scatter_pfold_over_mad,
    get_p2p_ssqr_diff_over_var,
)
from .scatter_res_raw import scatter_res_raw


__all__ = [
    "CADENCE_FEATS",
    "GENERAL_FEATS",
    "LOMB_SCARGLE_FEATS",
    "generate_dask_graph",
    "feature_categories",
    "dask_feature_graph",
]

feature_categories = {
    "Cadence/Error": [
        "n_epochs",
        "avg_err",
        "med_err",
        "std_err",
        "total_time",
        "avgt",
        "cads_std",
        "mean",
        "cads_avg",
        "cads_med",
        "cad_probs_1",
        "cad_probs_10",
        "cad_probs_20",
        "cad_probs_30",
        "cad_probs_40",
        "cad_probs_50",
        "cad_probs_100",
        "cad_probs_500",
        "cad_probs_1000",
        "cad_probs_5000",
        "cad_probs_10000",
        "cad_probs_50000",
        "cad_probs_100000",
        "cad_probs_500000",
        "cad_probs_1000000",
        "cad_probs_5000000",
        "cad_probs_10000000",
        "med_double_to_single_step",
        "avg_double_to_single_step",
        "std_double_to_single_step",
        "all_times_nhist_numpeaks",
        "all_times_nhist_peak_val",
        "all_times_nhist_peak_1_to_2",
        "all_times_nhist_peak_1_to_3",
        "all_times_nhist_peak_2_to_3",
        "all_times_nhist_peak_1_to_4",
        "all_times_nhist_peak_2_to_4",
        "all_times_nhist_peak_3_to_4",
        "all_times_nhist_peak1_bin",
        "all_times_nhist_peak2_bin",
        "all_times_nhist_peak3_bin",
        "all_times_nhist_peak4_bin",
    ],
    "General": [
        "amplitude",
        "flux_percentile_ratio_mid20",
        "flux_percentile_ratio_mid35",
        "flux_percentile_ratio_mid50",
        "flux_percentile_ratio_mid65",
        "flux_percentile_ratio_mid80",
        "max_slope",
        "maximum",
        "median",
        "median_absolute_deviation",
        "minimum",
        "percent_amplitude",
        "percent_beyond_1_std",
        "percent_close_to_median",
        "percent_difference_flux_percentile",
        "period_fast",
        "qso_log_chi2_qsonu",
        "qso_log_chi2nuNULL_chi2nu",
        "skew",
        "std",
        "stetson_j",
        "stetson_k",
        "weighted_average",
    ],
    "Lomb-Scargle (Periodic)": [
        "fold2P_slope_10percentile",
        "fold2P_slope_90percentile",
        "freq1_amplitude1",
        "freq1_amplitude2",
        "freq1_amplitude3",
        "freq1_amplitude4",
        "freq1_freq",
        "freq1_lambda",
        "freq1_rel_phase2",
        "freq1_rel_phase3",
        "freq1_rel_phase4",
        "freq1_signif",
        "freq2_amplitude1",
        "freq2_amplitude2",
        "freq2_amplitude3",
        "freq2_amplitude4",
        "freq2_freq",
        "freq2_rel_phase2",
        "freq2_rel_phase3",
        "freq2_rel_phase4",
        "freq3_amplitude1",
        "freq3_amplitude2",
        "freq3_amplitude3",
        "freq3_amplitude4",
        "freq3_freq",
        "freq3_rel_phase2",
        "freq3_rel_phase3",
        "freq3_rel_phase4",
        "freq_amplitude_ratio_21",
        "freq_amplitude_ratio_31",
        "freq_frequency_ratio_21",
        "freq_frequency_ratio_31",
        "freq_model_max_delta_mags",
        "freq_model_min_delta_mags",
        "freq_model_phi1_phi2",
        "freq_n_alias",
        "freq_signif_ratio_21",
        "freq_signif_ratio_31",
        "freq_varrat",
        "freq_y_offset",
        "linear_trend",
        "medperc90_2p_p",
        "p2p_scatter_2praw",
        "p2p_scatter_over_mad",
        "p2p_scatter_pfold_over_mad",
        "p2p_ssqr_diff_over_var",
        "scatter_res_raw",
    ],
}

CADENCE_FEATS = feature_categories["Cadence/Error"]
GENERAL_FEATS = feature_categories["General"]
LOMB_SCARGLE_FEATS = feature_categories["Lomb-Scargle (Periodic)"]


# See http://dask.pydata.org/en/latest/custom-graphs.html

dask_feature_graph = {
    "n_epochs": (len, "t"),
    "avg_err": (np.mean, "e"),
    "med_err": (np.median, "e"),
    "std_err": (np.std, "e"),
    "total_time": (lambda x: np.max(x) - np.min(x), "t"),
    "avgt": (np.mean, "t"),
    "cads": (np.diff, "t"),
    "cads_std": (np.std, "cads"),
    "mean": (np.mean, "m"),
    "cads_avg": (np.mean, "cads"),
    "cads_med": (np.median, "cads"),
    "cad_probs_1": (cad_prob, "cads", 1),
    "cad_probs_10": (cad_prob, "cads", 10),
    "cad_probs_20": (cad_prob, "cads", 20),
    "cad_probs_30": (cad_prob, "cads", 30),
    "cad_probs_40": (cad_prob, "cads", 40),
    "cad_probs_50": (cad_prob, "cads", 50),
    "cad_probs_100": (cad_prob, "cads", 100),
    "cad_probs_500": (cad_prob, "cads", 500),
    "cad_probs_1000": (cad_prob, "cads", 1000),
    "cad_probs_5000": (cad_prob, "cads", 5000),
    "cad_probs_10000": (cad_prob, "cads", 10000),
    "cad_probs_50000": (cad_prob, "cads", 50000),
    "cad_probs_100000": (cad_prob, "cads", 100000),
    "cad_probs_500000": (cad_prob, "cads", 500000),
    "cad_probs_1000000": (cad_prob, "cads", 1000000),
    "cad_probs_5000000": (cad_prob, "cads", 5000000),
    "cad_probs_10000000": (cad_prob, "cads", 10000000),
    "double_to_single_step": (double_to_single_step, "cads"),
    "avg_double_to_single_step": (np.mean, "double_to_single_step"),
    "med_double_to_single_step": (np.median, "double_to_single_step"),
    "std_double_to_single_step": (np.std, "double_to_single_step"),
    "delta_t_hist": (delta_t_hist, "t"),
    "delta_t_nhist": (normalize_hist, "delta_t_hist", "total_time"),
    "nhist_peaks": (find_sorted_peaks, "delta_t_nhist"),
    "all_times_nhist_numpeaks": (len, "nhist_peaks"),
    "all_times_nhist_peak_val": (np.max, "delta_t_nhist"),
    "all_times_nhist_peak_1_to_2": (peak_ratio, "nhist_peaks", 1, 2),
    "all_times_nhist_peak_1_to_3": (peak_ratio, "nhist_peaks", 1, 3),
    "all_times_nhist_peak_2_to_3": (peak_ratio, "nhist_peaks", 2, 3),
    "all_times_nhist_peak_1_to_4": (peak_ratio, "nhist_peaks", 1, 4),
    "all_times_nhist_peak_2_to_4": (peak_ratio, "nhist_peaks", 2, 4),
    "all_times_nhist_peak_3_to_4": (peak_ratio, "nhist_peaks", 3, 4),
    "all_times_nhist_peak1_bin": (peak_bin, "nhist_peaks", 1),
    "all_times_nhist_peak2_bin": (peak_bin, "nhist_peaks", 2),
    "all_times_nhist_peak3_bin": (peak_bin, "nhist_peaks", 3),
    "all_times_nhist_peak4_bin": (peak_bin, "nhist_peaks", 4),
    # Standalone features (disconnected nodes)
    "amplitude": (amplitude, "m"),
    "flux_percentile_ratio_mid20": (flux_percentile_ratio, "m", 20),
    "flux_percentile_ratio_mid35": (flux_percentile_ratio, "m", 35),
    "flux_percentile_ratio_mid50": (flux_percentile_ratio, "m", 50),
    "flux_percentile_ratio_mid65": (flux_percentile_ratio, "m", 65),
    "flux_percentile_ratio_mid80": (flux_percentile_ratio, "m", 80),
    "maximum": (maximum, "m"),
    "max_slope": (max_slope, "t", "m"),
    "median": (median, "m"),
    "median_absolute_deviation": (median_absolute_deviation, "m"),
    "minimum": (minimum, "m"),
    "percent_amplitude": (percent_amplitude, "m"),
    "percent_beyond_1_std": (percent_beyond_1_std, "m", "e"),
    "percent_close_to_median": (percent_close_to_median, "m"),
    "percent_difference_flux_percentile": (percent_difference_flux_percentile, "m"),
    "skew": (skew, "m"),
    "std": (std, "m"),
    "stetson_j": (stetson_j, "m"),
    "stetson_k": (stetson_k, "m"),
    "weighted_average": (weighted_average, "m", "e"),
    # QSO model features
    "qso_model": (qso_fit, "t", "m", "e"),
    "qso_log_chi2_qsonu": (get_qso_log_chi2_qsonu, "qso_model"),
    "qso_log_chi2nuNULL_chi2nu": (get_qso_log_chi2nuNULL_chi2nu, "qso_model"),
    # Fast Lomb-Scargle from Gatspy
    "period_fast": (lomb_scargle_fast_period, "t", "m", "e"),
    "_lomb_model": (lomb_scargle_model, "t", "m", "e"),
    # These could easily be programmatically generated, but this is more readable
    "freq1_freq": (get_lomb_frequency, "_lomb_model", 1),
    "freq2_freq": (get_lomb_frequency, "_lomb_model", 2),
    "freq3_freq": (get_lomb_frequency, "_lomb_model", 3),
    "freq1_amplitude1": (get_lomb_amplitude, "_lomb_model", 1, 1),
    "freq1_amplitude2": (get_lomb_amplitude, "_lomb_model", 1, 2),
    "freq1_amplitude3": (get_lomb_amplitude, "_lomb_model", 1, 3),
    "freq1_amplitude4": (get_lomb_amplitude, "_lomb_model", 1, 4),
    "freq2_amplitude1": (get_lomb_amplitude, "_lomb_model", 2, 1),
    "freq2_amplitude2": (get_lomb_amplitude, "_lomb_model", 2, 2),
    "freq2_amplitude3": (get_lomb_amplitude, "_lomb_model", 2, 3),
    "freq2_amplitude4": (get_lomb_amplitude, "_lomb_model", 2, 4),
    "freq3_amplitude1": (get_lomb_amplitude, "_lomb_model", 3, 1),
    "freq3_amplitude2": (get_lomb_amplitude, "_lomb_model", 3, 2),
    "freq3_amplitude3": (get_lomb_amplitude, "_lomb_model", 3, 3),
    "freq3_amplitude4": (get_lomb_amplitude, "_lomb_model", 3, 4),
    #        'freq1_rel_phase1': (get_lomb_rel_phase, 'lomb_model', 1, 1),
    "freq1_rel_phase2": (get_lomb_rel_phase, "_lomb_model", 1, 2),
    "freq1_rel_phase3": (get_lomb_rel_phase, "_lomb_model", 1, 3),
    "freq1_rel_phase4": (get_lomb_rel_phase, "_lomb_model", 1, 4),
    #        'freq2_rel_phase1': (get_lomb_rel_phase, 'lomb_model', 2, 1),
    "freq2_rel_phase2": (get_lomb_rel_phase, "_lomb_model", 2, 2),
    "freq2_rel_phase3": (get_lomb_rel_phase, "_lomb_model", 2, 3),
    "freq2_rel_phase4": (get_lomb_rel_phase, "_lomb_model", 2, 4),
    #        'freq3_rel_phase1': (get_lomb_rel_phase, 'lomb_model', 3, 1),
    "freq3_rel_phase2": (get_lomb_rel_phase, "_lomb_model", 3, 2),
    "freq3_rel_phase3": (get_lomb_rel_phase, "_lomb_model", 3, 3),
    "freq3_rel_phase4": (get_lomb_rel_phase, "_lomb_model", 3, 4),
    "freq_amplitude_ratio_21": (get_lomb_amplitude_ratio, "_lomb_model", 2),
    "freq_amplitude_ratio_31": (get_lomb_amplitude_ratio, "_lomb_model", 3),
    "freq_frequency_ratio_21": (get_lomb_frequency_ratio, "_lomb_model", 2),
    "freq_frequency_ratio_31": (get_lomb_frequency_ratio, "_lomb_model", 3),
    "freq_signif_ratio_21": (get_lomb_signif_ratio, "_lomb_model", 2),
    "freq_signif_ratio_31": (get_lomb_signif_ratio, "_lomb_model", 3),
    "freq1_lambda": (get_lomb_lambda, "_lomb_model"),
    "freq1_signif": (get_lomb_signif, "_lomb_model"),
    "freq_varrat": (get_lomb_varrat, "_lomb_model"),
    "linear_trend": (get_lomb_trend, "_lomb_model"),
    "freq_y_offset": (get_lomb_y_offset, "_lomb_model"),
    # Other features that operate on Lomb-Scargle residuals
    "freq_n_alias": (num_alias, "_lomb_model"),
    "scatter_res_raw": (scatter_res_raw, "t", "m", "e", "_lomb_model"),
    "_periodic_model": (periodic_model, "_lomb_model"),
    "_period_folded_model": (period_folding, "t", "m", "e", "_lomb_model"),
    "freq_model_max_delta_mags": (get_max_delta_mags, "_periodic_model"),
    "freq_model_min_delta_mags": (get_min_delta_mags, "_periodic_model"),
    "freq_model_phi1_phi2": (get_model_phi1_phi2, "_periodic_model"),
    "fold2P_slope_10percentile": (
        get_fold2P_slope_percentile,
        "_period_folded_model",
        10,
    ),
    "fold2P_slope_90percentile": (
        get_fold2P_slope_percentile,
        "_period_folded_model",
        90,
    ),
    "medperc90_2p_p": (get_medperc90_2p_p, "_period_folded_model"),
    "_p2p_model": (p2p_model, "t", "m", "freq1_freq"),
    "p2p_scatter_2praw": (get_p2p_scatter_2praw, "_p2p_model"),
    "p2p_scatter_over_mad": (get_p2p_scatter_over_mad, "_p2p_model"),
    "p2p_scatter_pfold_over_mad": (get_p2p_scatter_pfold_over_mad, "_p2p_model"),
    "p2p_ssqr_diff_over_var": (get_p2p_ssqr_diff_over_var, "_p2p_model"),
}


def generate_dask_graph(t, m, e):
    full_graph = {"t": t, "m": m, "e": e}
    full_graph.update(dask_feature_graph)
    return full_graph


extra_feature_docs = {
    "n_epochs": "Total number of observed values.",
    "avg_err": "Mean of the error estimates.",
    "med_err": "Median of error estimates.",
    "std_err": "Standard deviation of the error estimates.",
    "total_time": "Absolute difference between max and min of time values.",
    "avgt": "Mean of the time values.",
    "mean": "Mean of observed values.",
    "cads": "List of differences between successive time values (`np.diff(t)`).",
    "cads_std": "Standard deviation of `cads` (discrete difference between times).",
    "cads_avg": "Mean value of `cads` (discrete difference between times).",
    "cads_med": "Median value of `cads` (discrete difference between times).",
    "avg_double_to_single_step": "Mean value of ratios (t[i+2] - t[i]) / (t[i+2] - t[i+1]).",
    "med_double_to_single_step": "Median value of ratios (t[i+2] - t[i]) / (t[i+2] - t[i+1]).",
    "std_double_to_single_step": "Standard deviation of ratios (t[i+2] - t[i]) / (t[i+2] - t[i+1]).",
    "all_times_nhist_numpeaks": "Number of peaks (local maxima) in histogram of all possible delta_t's.",
    "all_times_nhist_peak_val": "Peak value in histogram of all possible delta_t's.",
}


feature_tags = {
    "n_epochs": ["Astronomy", "General"],
    "avg_err": ["Astronomy", "Error", "General"],
    "med_err": ["Astronomy", "Error", "General"],
    "std_err": ["Astronomy", "Error", "General"],
    "total_time": ["Astronomy", "General"],
    "avgt": ["Astronomy", "General"],
    "cads": ["Astronomy", "General", "Cadence"],
    "cads_std": ["Astronomy", "General", "Cadence"],
    "mean": ["Astronomy", "General"],
    "cads_avg": ["Astronomy", "General", "Cadence"],
    "cads_med": ["Astronomy", "General", "Cadence"],
    "cad_probs_1": ["Astronomy", "General", "Cadence"],
    "cad_probs_10": ["Astronomy", "General", "Cadence"],
    "cad_probs_20": ["Astronomy", "General", "Cadence"],
    "cad_probs_30": ["Astronomy", "General", "Cadence"],
    "cad_probs_40": ["Astronomy", "General", "Cadence"],
    "cad_probs_50": ["Astronomy", "General", "Cadence"],
    "cad_probs_100": ["Astronomy", "General", "Cadence"],
    "cad_probs_500": ["Astronomy", "General", "Cadence"],
    "cad_probs_1000": ["Astronomy", "General", "Cadence"],
    "cad_probs_5000": ["Astronomy", "General", "Cadence"],
    "cad_probs_10000": ["Astronomy", "General", "Cadence"],
    "cad_probs_50000": ["Astronomy", "General", "Cadence"],
    "cad_probs_100000": ["Astronomy", "General", "Cadence"],
    "cad_probs_500000": ["Astronomy", "General", "Cadence"],
    "cad_probs_1000000": ["Astronomy", "General", "Cadence"],
    "cad_probs_5000000": ["Astronomy", "General", "Cadence"],
    "cad_probs_10000000": ["Astronomy", "General", "Cadence"],
    "double_to_single_step": ["Astronomy", "General", "Cadence"],
    "avg_double_to_single_step": ["Astronomy", "General", "Cadence"],
    "med_double_to_single_step": ["Astronomy", "General", "Cadence"],
    "std_double_to_single_step": ["Astronomy", "General", "Cadence"],
    "delta_t_hist": ["Astronomy", "General", "Cadence"],
    "delta_t_nhist": ["Astronomy", "General", "Cadence"],
    "nhist_peaks": ["Astronomy", "General", "Cadence"],
    "all_times_nhist_numpeaks": ["Astronomy", "General", "Cadence"],
    "all_times_nhist_peak_val": ["Astronomy", "General", "Cadence"],
    "all_times_nhist_peak_1_to_2": ["Astronomy", "General", "Cadence"],
    "all_times_nhist_peak_1_to_3": ["Astronomy", "General", "Cadence"],
    "all_times_nhist_peak_2_to_3": ["Astronomy", "General", "Cadence"],
    "all_times_nhist_peak_1_to_4": ["Astronomy", "General", "Cadence"],
    "all_times_nhist_peak_2_to_4": ["Astronomy", "General", "Cadence"],
    "all_times_nhist_peak_3_to_4": ["Astronomy", "General", "Cadence"],
    "all_times_nhist_peak1_bin": ["Astronomy", "General", "Cadence"],
    "all_times_nhist_peak2_bin": ["Astronomy", "General", "Cadence"],
    "all_times_nhist_peak3_bin": ["Astronomy", "General", "Cadence"],
    "all_times_nhist_peak4_bin": ["Astronomy", "General", "Cadence"],
    # Standalone features (disconnected nodes)
    "amplitude": ["Astronomy", "General"],
    "flux_percentile_ratio_mid20": ["Astronomy"],
    "flux_percentile_ratio_mid35": ["Astronomy"],
    "flux_percentile_ratio_mid50": ["Astronomy"],
    "flux_percentile_ratio_mid65": ["Astronomy"],
    "flux_percentile_ratio_mid80": ["Astronomy"],
    "maximum": ["Astronomy", "General"],
    "max_slope": ["Astronomy", "General"],
    "median": ["Astronomy", "General"],
    "median_absolute_deviation": ["Astronomy", "General"],
    "minimum": ["Astronomy", "General"],
    "percent_amplitude": ["Astronomy", "General"],
    "percent_beyond_1_std": ["Astronomy", "General"],
    "percent_close_to_median": ["Astronomy", "General"],
    "percent_difference_flux_percentile": ["Astronomy", "General"],
    "skew": ["Astronomy", "General"],
    "std": ["Astronomy", "General"],
    "stetson_j": ["Astronomy", "General"],
    "stetson_k": ["Astronomy", "General"],
    "weighted_average": ["Astronomy", "General"],
    # QSO model features
    "qso_model": ["Astronomy"],
    "qso_log_chi2_qsonu": ["Astronomy"],
    "qso_log_chi2nuNULL_chi2nu": ["Astronomy"],
    # Fast Lomb-Scargle from Gatspy
    "period_fast": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "_lomb_model": ["Astronomy", "Periodic", "Lomb-Scargle"],
    # These could easily be programmatically generated, but this is more readable
    "freq1_freq": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq2_freq": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq3_freq": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq1_amplitude1": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq1_amplitude2": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq1_amplitude3": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq1_amplitude4": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq2_amplitude1": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq2_amplitude2": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq2_amplitude3": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq2_amplitude4": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq3_amplitude1": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq3_amplitude2": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq3_amplitude3": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq3_amplitude4": ["Astronomy", "Periodic", "Lomb-Scargle"],
    #        'freq1_rel_phase1': (get_lomb_rel_phase, 'lomb_model', 1, 1),
    "freq1_rel_phase2": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq1_rel_phase3": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq1_rel_phase4": ["Astronomy", "Periodic", "Lomb-Scargle"],
    #        'freq2_rel_phase1': (get_lomb_rel_phase, 'lomb_model', 2, 1),
    "freq2_rel_phase2": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq2_rel_phase3": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq2_rel_phase4": ["Astronomy", "Periodic", "Lomb-Scargle"],
    #        'freq3_rel_phase1': (get_lomb_rel_phase, 'lomb_model', 3, 1),
    "freq3_rel_phase2": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq3_rel_phase3": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq3_rel_phase4": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq_amplitude_ratio_21": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq_amplitude_ratio_31": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq_frequency_ratio_21": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq_frequency_ratio_31": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq_signif_ratio_21": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq_signif_ratio_31": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq1_lambda": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq1_signif": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq_varrat": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "linear_trend": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq_y_offset": ["Astronomy", "Periodic", "Lomb-Scargle"],
    # Other features that operate on Lomb-Scargle residuals
    "freq_n_alias": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "scatter_res_raw": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "_periodic_model": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "_period_folded_model": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq_model_max_delta_mags": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq_model_min_delta_mags": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "freq_model_phi1_phi2": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "fold2P_slope_10percentile": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "fold2P_slope_90percentile": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "medperc90_2p_p": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "_p2p_model": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "p2p_scatter_2praw": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "p2p_scatter_over_mad": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "p2p_scatter_pfold_over_mad": ["Astronomy", "Periodic", "Lomb-Scargle"],
    "p2p_ssqr_diff_over_var": ["Astronomy", "Periodic", "Lomb-Scargle"],
}



import os
import tarfile
from zipfile import ZipFile

import numpy as np
import numpy.testing as npt
import pandas as pd

from cesium.datasets import andrzejak
from cesium.datasets import asas_training

from io import BytesIO as StringIO

try:
    import mock
except ImportError:
    import unittest.mock as mock


def urlpatch(*args):
    import urllib.request  # noqa: F401

    return mock.patch("urllib.request.urlopen", side_effect=args[0])


class MockZipUrl(object):
    def __init__(self, url, num_columns=1):
        self.url = url
        self.num_columns = num_columns

    def read(self):
        out_buffer = StringIO()
        fname = os.path.basename(self.url)
        prefix = fname[0]
        if fname.endswith(".dat"):
            header = "filename,class,meta1\n{}001,{},1.0\n".format(prefix, prefix)
            out_buffer.write(header.encode("utf-8"))
        else:
            X = np.random.random((10, self.num_columns))
            text = "\n".join(",".join(row) for row in X.astype("str"))
            if fname.endswith(".zip"):
                with ZipFile(out_buffer, "w") as z:
                    z.writestr("{}001.txt".format(prefix), text)
            elif fname.endswith(".tar.gz"):
                with tarfile.open(fileobj=out_buffer, mode="w:gz") as t:
                    tarinfo = tarfile.TarInfo("{}001.txt".format(prefix))
                    tarinfo.size = len(text)
                    t.addfile(tarinfo, StringIO(text.encode("utf-8")))
        return out_buffer.getvalue()


def _mock_andrzejak_url(*args):
    return MockZipUrl(url=args[0], num_columns=1)


@urlpatch(_mock_andrzejak_url)
def test_fetch_andrzejak(self, tmpdir):
    """Test EEG data download."""
    num_files = len(andrzejak.ZIP_FILES)
    andrzejak.MD5SUMS = None
    data = andrzejak.fetch_andrzejak(str(tmpdir))
    assert data["archive"].endswith("andrzejak.tar.gz") and os.path.exists(
        data["archive"]
    )
    header = pd.read_csv(data["header"])
    npt.assert_array_equal(header["class"], data["classes"])
    assert all(len(t) == len(m) for t, m in zip(data["times"], data["measurements"]))
    assert len(data["times"]) == num_files
    assert len(data["classes"]) == num_files
    assert all(c in ["Z", "N", "O", "S", "F"] for c in data["classes"])


def _mock_asas_training_url(*args):
    return MockZipUrl(url=args[0], num_columns=3)


@urlpatch(_mock_asas_training_url)
def test_fetch_asas_training(self, tmpdir):
    """Test ASAS training data download."""
    num_files = 1
    asas_training.MD5SUMS = None
    data = asas_training.fetch_asas_training(str(tmpdir))
    assert data["archive"].endswith("asas_training_set.tar.gz") and os.path.exists(
        data["archive"]
    )
    header = pd.read_csv(data["header"])
    npt.assert_array_equal(header["class"], data["classes"])
    assert all(
        len(t) == len(m) and len(m) == len(e)
        for t, m, e in zip(data["times"], data["measurements"], data["errors"])
    )
    assert len(data["times"]) == num_files
    assert len(data["classes"]) == num_files
    assert len(data["metadata"]) == num_files
    assert all(c in ["a"] for c in data["classes"])



#! /usr/bin/env python

import os
import sys

import setuptools
from distutils.command.build_py import build_py


descr = """Machine Learning Time Series Platform

https://github.com/cesium-ml/cesium

"""

DISTNAME = "cesium"
DESCRIPTION = "Machine Learning Time-Series Platform"
LONG_DESCRIPTION = descr
MAINTAINER = "cesium Team"
MAINTAINER_EMAIL = "stefanv@berkeley.edu"
URL = "http://cesium-ml.org"
LICENSE = "Modified BSD"
DOWNLOAD_URL = "https://github.com/cesium-ml/cesium"
PYTHON_VERSION = (3, 4)


def configuration(parent_package="", top_path=None):
    if os.path.exists("MANIFEST"):
        os.remove("MANIFEST")

    from numpy.distutils.misc_util import Configuration

    config = Configuration(None, parent_package, top_path)

    config.set_options(
        ignore_setup_xxx_py=True,
        assume_default_configuration=True,
        delegate_options_to_subpackages=True,
        quiet=True,
    )

    config.add_subpackage("cesium")

    return config


with open("cesium/version.py") as fid:
    for line in fid:
        if line.startswith("version"):
            VERSION = line.split("=")[-1].strip(" '\"\n")
            break

with open("requirements.txt") as fid:
    INSTALL_REQUIRES = [
        line.split("#")[0].strip()
        for line in fid.readlines()
        if line and not line.startswith("git")
    ]
    INSTALL_REQUIRES = [pkg.replace("-", "_") for pkg in INSTALL_REQUIRES]


# requirements for those browsing PyPI
REQUIRES = [r.replace(">=", " (>= ") + ")" if "=" in r else r for r in INSTALL_REQUIRES]
REQUIRES = [r.replace("==", " (== ") for r in REQUIRES]


if __name__ == "__main__":
    try:
        from numpy.distutils.core import setup

        extra = {"configuration": configuration}
        # do not risk updating numpy
        INSTALL_REQUIRES = [r for r in INSTALL_REQUIRES if "numpy" not in r]
    except ImportError:
        if len(sys.argv) >= 2 and (
            "--help" in sys.argv[1:]
            or sys.argv[1] in ("--help-commands", "egg_info", "--version", "clean")
        ):
            # For these actions, NumPy is not required.
            #
            # They are required to succeed without Numpy for example when
            # pip is used to install cesium when Numpy is not yet
            # present in the system.
            try:
                from setuptools import setup
            except ImportError:
                from distutils.core import setup
        else:
            print(
                "To install cesium from source, you will need numpy.\n"
                + "Install numpy with pip:\n"
                + "  pip install numpy\n"
                "Or using conda:\n"
                "  conda install numpy\n"
                "or use your operating system package manager."
            )
            sys.exit(1)

    setup(
        name=DISTNAME,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        url=URL,
        license=LICENSE,
        download_url=DOWNLOAD_URL,
        version=VERSION,
        classifiers=[
            "Development Status :: 4 - Beta",
            "Environment :: Console",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: BSD License",
            "Programming Language :: C",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: Unix",
            "Operating System :: MacOS",
        ],
        install_requires=INSTALL_REQUIRES,
        setup_requires=["cython>=0.25"],
        requires=REQUIRES,
        configuration=configuration,
        packages=setuptools.find_packages(exclude=["doc"]),
        include_package_data=True,
        zip_safe=False,
        cmdclass={"build_py": build_py},
    )



from .graphs import (
    CADENCE_FEATS,
    GENERAL_FEATS,
    LOMB_SCARGLE_FEATS,
    generate_dask_graph,
    feature_categories,
    dask_feature_graph,
    feature_tags,
)

__all__ = [
    CADENCE_FEATS,
    GENERAL_FEATS,
    LOMB_SCARGLE_FEATS,
    generate_dask_graph,
    feature_categories,
    dask_feature_graph,
    feature_tags,
]



import numpy as np
from scipy.stats import norm
from scipy.linalg import solveh_banded, cholesky_banded
from scipy.special import gammaln, betainc, gammaincc


# TODO duplicate
def lprob2sigma(lprob):
    """Translates a log_e(probability) to units of Gaussian sigmas."""
    if lprob > -36.0:
        sigma = norm.ppf(1.0 - 0.5 * np.exp(1.0 * lprob))
    else:
        sigma = np.sqrt(np.log(2.0 / np.pi) - 2.0 * np.log(8.2) - 2.0 * lprob)
    return float(sigma)


def chol_inverse_diag(t):
    """Computes inverse of matrix given its Cholesky upper Triangular decomposition t.
    matrix form: ab[u + i - j, j] == a[i,j] (here u=1)
    (quick version: only calculates diagonal and neighboring elements)
    """
    (uu, nrows) = t.shape
    B = np.zeros((uu, nrows), dtype="float64")
    B[1, nrows - 1] = 1.0 / t[1, nrows - 1] ** 2
    B[0, nrows - 1] = -t[0, nrows - 1] * B[1, nrows - 1] / t[1, nrows - 2]
    for j in reversed(range(nrows - 1)):
        tjj = t[1, j]
        B[1, j] = (1.0 / tjj - t[0, j + 1] * B[0, j + 1]) / tjj
        B[0, j] = -t[0, j] * B[1, j] / t[1, j - 1]
    return B


def qso_engine(time, data, error, ltau=3.0, lvar=-1.7, sys_err=0.0, return_model=False):
    """Calculates the fit quality of a damped random walk to a qso lightcurve.
    The formalism is from Rybicki & Press (1994; arXiv:comp-gas/9405004)

    Data are modelled with a covariance function
        Lij = 0.5*var*tau*exp(-|time_i-time_j|/tau) .

    Input:
        time - measurement times, typically days
        data - measured magnitudes
        error - uncertainty in measured magnitudes

    Output (dictionary):

        chi2/nu - classical variability measure
        chi2_qso/nu - for goodness of fit given fixed parameters
        chi2_qso/nu_extra - for parameter fitting, add to chi2/nu
        chi^2/nu_NULL - expected chi2/nu for non-qso variable

        signif_qso - significance chi^2/nu<chi^2/nu_NULL (rule out false alarm)
        signif_not_qso - significance chi^2/nu>1 (rule out qso)
        signif_vary - significance that source is variable
        class - resulting source type (ambiguous, not_qso, qso)

        model - time series prediction for each datum given all others (iff return_model==True)
        dmodel - model uncertainty, including uncertainty in data

    Notes:
        T = L^(-1)
        Data variance is D
        Full covariance C^(-1) = (L+D)^(-1) = T [T+D^(-1)]^(-1) D^(-1)
        Code takes advantage of the tridiagonality of T and T+D^(-1).
    """

    out_dict = {}
    out_dict["chi2_qso/nu"] = 999
    out_dict["chi2_qso/nu_extra"] = 0.0
    out_dict["signif_qso"] = 0.0
    out_dict["signif_not_qso"] = 0.0
    out_dict["signif_vary"] = 0.0
    out_dict["chi2_qso/nu_NULL"] = 0.0
    out_dict["chi2/nu"] = 0.0
    out_dict["nu"] = 0
    out_dict["model"] = []
    out_dict["dmodel"] = []
    out_dict["class"] = "ambiguous"

    lvar0 = np.log10(0.5) + lvar + ltau

    ln = len(data)
    dt = abs(time[1:] - time[:-1])

    # first make sure all dt>0
    g = np.where(dt > 0.0)[0]
    lg = len(g)
    # must have at least 2 data points
    if lg <= 0:
        return out_dict

    if return_model:
        model = 1.0 * data
        dmodel = -1.0 * error

    if lg < ln:
        dt = dt[g]
        gg = np.zeros(lg + 1, dtype="int64")
        gg[1:] = g + 1
        dat = data[gg]
        wt = 1.0 / (sys_err**2 + error[gg] ** 2)
        ln = lg + 1
    else:
        dat = 1.0 * data
        wt = 1.0 / (sys_err**2 + error**2)

    out_dict["nu"] = ln - 1.0
    varx = np.var(dat)
    dat0 = (dat * wt).sum() / wt.sum()
    out_dict["chi2/nu"] = ((dat - dat0) ** 2 * wt).sum() / out_dict["nu"]

    # define tridiagonal matrix T = L^(-1)
    # sparse matrix form: ab[u + i - j, j] == a[i,j]   i<=j, (here u=1)
    T = np.zeros((2, ln), dtype="float64")
    arg = dt * np.exp(-np.log(10) * ltau)
    ri = np.exp(-arg)
    ei = 1.0 / (1.0 / ri - ri)
    T[0, 1:] = -ei
    T[1, :-1] = 1.0 + ri * ei
    T[1, 1:] += ri * ei
    T[1, ln - 1] += 1.0
    T0 = np.median(T[1, :])
    T /= T0

    # equation for chi2_qso is [ (dat-x0) T Tp^(-1) D^(-1) (dat-x0) ]  , where Tp=T+D^(-1) and D^(-1)=wt
    fac = np.exp(np.log(10) * lvar0) / T0
    Tp = 1.0 * T
    Tp[1, :] += wt * fac
    # solve Tp*z=y for z (y=wt*dat)
    # This works for scipy __version__>='0.9.0' on anathem (20120809)
    b1 = (wt * dat).reshape((1, ln))
    b2 = b1.T
    # (Tpc,z) = solveh_banded(Tp,b2)
    z = solveh_banded(Tp, b2)
    Tpc = cholesky_banded(
        Tp
    )  # the solveh_banded() function used to return the cholesky matrix, now we get seperately
    z = z.T
    z = z[0, :]
    c1 = wt.reshape((1, ln))
    c2 = c1.T
    # (Tpc,z0) = solveh_banded(Tp,c2)
    z0 = solveh_banded(Tp, c2)
    # HAS NOT CHANGED#Tpc2 = cholesky_banded(Tp)
    z0 = z0.T
    z0 = z0[0, :]

    # finally, get u=T*z
    u = T[1, :] * z
    u[1:] += T[0, 1:] * z[:-1]
    u[:-1] += T[0, 1:] * z[1:]
    u0 = T[1, :] * z0
    u0[1:] += T[0, 1:] * z0[:-1]
    u0[:-1] += T[0, 1:] * z0[1:]

    # magnitude offset x0, error = 1./sqrt(u0sum)
    u0sum = u0.sum()
    x0 = u.sum() / u0sum

    # fit statistic
    out_dict["chi2_qso/nu"] = np.dot(dat - x0, u - u0 * x0) / out_dict["nu"]

    # -2*log(likelihood) = chi2_qso + ldet_C + log(u0sum)
    #   first term: use chi2_qso/nu for goodness of fit with fixed parameters;
    #   all terms: use chi2_qso/nu + chi2_qso/nu_extra for fitting with variable parameters
    # get log of determinant for use later
    Tc = cholesky_banded(T)
    ldet_Tp = 2 * np.log(Tpc[1, :]).sum()
    ldet_T = 2 * np.log(Tc[1, :]).sum()
    ldet_C = ldet_Tp - ldet_T - np.log(wt).sum()
    out_dict["chi2_qso/nu_extra"] = (ldet_C + np.log(u0sum)) / out_dict["nu"]

    # get trace of C^(-1) for significance calculation
    Tpm = chol_inverse_diag(Tpc)
    diagC = T[1, :] * wt * Tpm[1, :]
    diagC[:-1] += T[0, 1:] * wt[0:-1] * Tpm[0, 1:]
    diagC[1:] += T[0, 1:] * wt[1:] * Tpm[0, 1:]
    TrC = diagC.sum()

    # significance in sigma units (large means false alarm unlikely)
    # (expected value of chi2_qso under the NULL hypothesis is TrC*varx)
    out_dict["chi2_qso/nu_NULL"] = TrC * varx / out_dict["nu"]
    a = ln / 2.0
    x = (out_dict["chi2_qso/nu"] + 1.0e-8) / (
        out_dict["chi2_qso/nu_NULL"] + out_dict["chi2_qso/nu"] + 1.0e-8
    )
    prob = betainc(a, a, x)
    if prob <= 0:
        lprob = a * np.log(x) - np.log(a) + gammaln(2 * a) - 2 * gammaln(a)
    else:
        lprob = np.log(prob)
    out_dict["signif_qso"] = lprob2sigma(lprob)

    a = ln / 2.0
    x = 1.0 / (1.0 + out_dict["chi2_qso/nu"])
    prob = betainc(a, a, x)
    if prob <= 0:
        lprob = a * np.log(x) - np.log(a) + gammaln(2 * a) - 2 * gammaln(a)
    else:
        lprob = np.log(prob)
    out_dict["signif_not_qso"] = lprob2sigma(lprob)

    x = out_dict["chi2/nu"] * out_dict["nu"]
    prob = gammaincc(0.5 * out_dict["nu"], 0.5 * x)
    if prob <= 0:
        lprob = (
            (0.5 * out_dict["nu"] - 1) * np.log(x)
            - 0.5 * x
            - 0.5 * out_dict["nu"] * np.log(2)
            - gammaln(0.5 * out_dict["nu"])
        )
    else:
        lprob = np.log(prob)
    out_dict["signif_vary"] = lprob2sigma(lprob)

    if out_dict["signif_vary"] > 3:
        if out_dict["signif_qso"] > 3:
            out_dict["class"] = "qso"
        elif out_dict["signif_not_qso"] > 3:
            out_dict["class"] = "not_qso"

    # best-fit model for the lightcurve
    if return_model:
        model[gg] = dat - (u - u0 * x0) / diagC
        dmodel[gg] = 1.0 / np.sqrt(diagC)
        out_dict["model"] = model
        out_dict["dmodel"] = dmodel

    return out_dict


def qso_fit(time, data, error, filter="g", mag0=19.0, sys_err=0.0, return_model=False):
    """Best-fit qso model determined for Sesar Strip82, ugriz-bands (default r).
    See additional notes for underlying code qso_engine.

    Input:
        time - measurement times [days]
        data - measured magnitudes in single filter (also specified)
        error - uncertainty in measured magnitudes

    Output:
        chi^2/nu - classical variability measure
        chi^2_qso/nu - fit statistic
        chi^2_qso/nu_NULL - expected fit statistic for non-qso variable

        signif_qso - significance chi^2/nu<chi^2/nu_NULL (rule out false alarm)
        signif_not_qso - significance chi^2/nu>1 (rule out qso)
        signif_vary - significance that source is variable at all
        class - source type (ambiguous, not_qso, qso)

        model - time series prediction for each datum given all others (iff return_model==True)
        dmodel - model uncertainty, including uncertainty in data

    Note on use (i.e., how class is defined):

          (0) signif_vary < 3: ambiguous, else
          (1) signif_qso > 3: qso, else
          (2) signif_not_qso > 3: not_qso
    """

    data = data.copy() - np.median(data) + mag0
    pars = {}
    pars["u"] = [-3.90, 0.12, 2.73, -0.02]
    pars["g"] = [-4.10, 0.14, 2.92, -0.07]
    pars["r"] = [-4.34, 0.20, 3.12, -0.15]
    pars["i"] = [-4.23, 0.05, 2.83, 0.07]
    pars["z"] = [-4.44, 0.13, 3.06, -0.07]

    par = pars[filter.lower()]
    lvar = par[0] + par[1] * (mag0 - 19.0)
    ltau = par[2] + par[3] * (mag0 - 19.0)

    adict = qso_engine(
        time,
        data,
        error,
        ltau=ltau,
        lvar=lvar,
        return_model=return_model,
        sys_err=sys_err,
    )

    out_dict = {}
    out_dict["lvar"] = lvar
    out_dict["ltau"] = ltau
    out_dict["chi2/nu"] = adict["chi2/nu"]
    out_dict["nu"] = adict["nu"]
    out_dict["chi2_qso/nu"] = adict["chi2_qso/nu"]
    out_dict["chi2_qso/nu_NULL"] = adict["chi2_qso/nu_NULL"]
    out_dict["signif_qso"] = adict["signif_qso"]
    out_dict["signif_not_qso"] = adict["signif_not_qso"]
    out_dict["signif_vary"] = adict["signif_vary"]
    out_dict["class"] = adict["class"]
    out_dict["chi2qso_nu_nuNULL_ratio"] = (
        out_dict["chi2_qso/nu"] / out_dict["chi2_qso/nu_NULL"]
    )

    # ---
    # Nat has converged upon the following being the most significant features,
    # Joey believes it is best to just use these features only (so now the others are disabled in
    # __init__.py and qso_extractor.py
    out_dict["log_chi2_qsonu"] = np.log(out_dict["chi2_qso/nu"])
    out_dict["log_chi2nuNULL_chi2nu"] = np.log(
        out_dict["chi2_qso/nu_NULL"] / out_dict["chi2_qso/nu"]
    )
    # ---

    if return_model:
        out_dict["model"] = adict["model"]
        out_dict["dmodel"] = adict["dmodel"]

    return out_dict


def get_qso_log_chi2_qsonu(qso_model):
    """Natural log of goodness of fit of qso-model given fixed parameters."""
    return qso_model["log_chi2_qsonu"]


def get_qso_log_chi2nuNULL_chi2nu(qso_model):
    """Natural log of expected chi2/nu for non-qso variable."""
    return qso_model["log_chi2nuNULL_chi2nu"]



import numpy as np


# TODO is this really a useful feature?
def num_alias(lomb_model):
    """Here we check for "1-day" aliases in ASAS / Deboss sources."""
    alias = [
        {
            "per": 1.0,
            "p_low": 0.92,
            "p_high": 1.08,
            "alpha_1": 8.191855,
            "alpha_2": -7.976243,
        },
        {
            "per": 0.5,
            "p_low": 0.48,
            "p_high": 0.52,
            "alpha_1": 2.438913,
            "alpha_2": 0.9837243,
        },
        {
            "per": 0.3333333333,
            "p_low": 0.325,
            "p_high": 0.342,
            "alpha_1": 2.95749,
            "alpha_2": -4.285432,
        },
        {
            "per": 0.25,
            "p_low": 0.245,
            "p_high": 0.255,
            "alpha_1": 1.347657,
            "alpha_2": 2.326338,
        },
    ]

    count = 0
    for freq_dict in lomb_model["freq_fits"]:
        period = 1.0 / freq_dict["freq"]
        for a in alias:
            cutoff = (
                a["alpha_1"] / np.power(np.abs(period - a["per"]), 0.25) + a["alpha_2"]
            )
            if (
                period >= a["p_low"]
                and period <= a["p_high"]
                and freq_dict["signif"] < cutoff
            ):
                count += 1  # this frequency has a "1 day" alias (or 0.5 or 0.33
                break  # only need to do this once per period, if an alias is found.
    return count



import os
import numpy as np
from Cython.Build import cythonize

base_path = os.path.abspath(os.path.dirname(__file__))


def configuration(parent_package="", top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration("features", parent_package, top_path)

    cythonize(os.path.join(base_path, "_lomb_scargle.pyx"))

    config.add_extension(
        "_lomb_scargle", "_lomb_scargle.c", include_dirs=[np.get_include()]
    )

    return config


if __name__ == "__main__":
    from numpy.distutils.core import setup

    setup(**configuration(top_path="").todict())
